' The following specs are supported by all DSP versions (this means
' all SoundBlaster cards can handle these settings:)
'
'   Transfer: 8-bit Mono PCM Single-cycle
'       Wave: 8-bit Mono 11025hz uncompressed

'$INCLUDE: 'c:\QB45\QB.BI'
'$INCLUDE: 'AUDLIB.BI'

DEFINT A-Z

CONST cardSB1 = 1     ' Sound Blaster 1.0/1.5
CONST cardSBPro = 2   ' Sound Blaster Pro
CONST cardSB2 = 3     ' Sound Blaster 2.0
CONST cardSBPro2 = 4  ' Sound Blaster Pro 2
CONST cardSBPro2b = 5 ' Sound Blaster Pro 2 (+FM Synth)
CONST cardSB16 = 6    ' Sound Blaster 16/32/AWE32/AWE64

CONST chipCT1335 = 1  ' Mono only, doesn't support bass & treble (SB2CD)
CONST chipCT1345 = 2  ' Stereo, doesn't support bass & treble (Sound Blaster Pro)
CONST chipCT1745 = 3  ' Sound Blaster 16 and more recent

TYPE sbConfig         ' OFS SZE   DESCRIPTION
	port AS INTEGER     ' 000 ..2   SB card base I/O port
	irq AS INTEGER      ' 002 ..2   Interrupt request line
	loDMA AS INTEGER    ' 004 ..2   8-bit DMA channel
	HiDMA AS INTEGER    ' 006 ..2   16-bit DMA channel
	mixer AS INTEGER    ' 008 ..2   Mixer base I/O port
	mpu AS INTEGER      ' 010 ..2   MPU-401 base I/O port
	series AS INTEGER   ' 012 ..2   Card family
	dspVer AS INTEGER   ' 014 ..2   DSP version
	chip AS INTEGER     ' 016 ..2   Mixer chipset

	DMAMask AS INTEGER  ' 018 ..2   DMA Mask port
	DMAMode AS INTEGER  ' 020 ..2   DMA Mode port
	DMAClr AS INTEGER   ' 022 ..2   DMA Clear Register port
	DMAStat AS INTEGER  ' 024 ..2   DMA Status Register port
	DMAAddr AS INTEGER  ' 026 ..2   DMA Address port
	DMASize AS INTEGER  ' 028 ..2   DMA Size port
	DMAPage AS INTEGER  ' 030 ..2   DMA Page port
END TYPE              ' 32 BYTES

DIM SHARED sb AS sbConfig
''
'' Note: the mixer is reinitialized on startup. I think it should
'' keep track of the initial values before reinitialization, and then
'' restore those values before quitting. I might need a audShutDown
'' routine for that.
''

'' Mixer chipsets:
'' CT1335 (mono) - found in SB1.0, SB1.5, SB2.0 CD
''   MASTER (3 bits), MIDI (3 bits), CD (3 bits), VOICE (3 bits)
''
'' CT1345 (stereo) - found in SBPro, SBPro2, SBPro2 + FM Synth
''  MASTER (3 bits), MIDI (3 bits), CD (3 bits), VOICE (3 bits)
''  Supports low-pass filter output.
''
'' CT1745 (stereo) - found in SB16
''  MASTER (5 bits), MIDI (5 bits), CD (5 bits), VOICE (5 bits)
''  Supports BASS (4 bits), TREBLE (4 bits)
''
'' The SB 2.0 has no mixer chip at all.

' DSP read/write/reset
DECLARE FUNCTION audDSPRead% ()
DECLARE SUB audDSPWrite (anyByte AS INTEGER)
DECLARE FUNCTION audDSPReset% ()

' Mixer read/write
DECLARE FUNCTION audMixerRead% (regIndex AS INTEGER)
DECLARE SUB audMixerWrite (regIndex AS INTEGER, anyByte AS INTEGER)
DECLARE SUB audVolSetGen (regIndex AS INTEGER, lf AS INTEGER, rt AS INTEGER)
DECLARE SUB audVolGetGen (regIndex AS INTEGER, lf AS INTEGER, rt AS INTEGER)

''
'' Returns TRUE if DMA transfer is over
''
FUNCTION audDMADone%
	DIM tmp AS INTEGER

	' The channel availability bit should be set if DMA transfer is over
	IF (INP(sb.DMAStat) AND (2 ^ (sb.loDMA AND &H3))) THEN
		' Read the DSP Read-Buffer Status port once to acknowledge the DSP
		' interrupt (should be done from a custom ISR, see IRQ parameter
		' in BLASTER environment string.)
		tmp = INP(sb.port + &HE)
		' Acknowledge the end of interrupt with the Programmable Interrupt
		' Controller (PIC); extra step required for IRQ 8-15.
		OUT &H20, &H20
		SELECT CASE sb.irq
			CASE 8 TO 15
				OUT &HA0, &H20
		END SELECT
		' Transfer is done
		audDMADone% = -1
	END IF
END FUNCTION

''
'' Make sure the DSP Buffer is loaded before attempting to read data
'' from it. The Read-Buffer Status bit 7 is clear if the DSP buffer is
'' empty, set if there's something there.
''
'' Valid for SB1.5, SBMCV, SB2.0, SBPRO, SBPRO MCV, SB16
''
FUNCTION audDSPRead%
	DO: LOOP UNTIL (INP(sb.port + &HE) AND 128) ' wait for "ready"
	audDSPRead% = INP(sb.port + &HA)            ' read data
END FUNCTION

''
'' The reset DSP command can be used to confirm the presence of a SB
'' at the specified port, to stop data transfer before the end of a
'' block, and to stop high-speed auto-initialize DMA mode.
'' First we send a reset request (write 1 to the reset port, wait 3
'' msec, then write 0 to the reset port.) Then we wait for the DSP to
'' confirm the reset (it should take 100 ms at most to respond.) Poll
'' the DSP Read-Buffer Status port and examine bit 7. The DSP buffer
'' is empty if bit 7 is clear; it contains something if bit 7 is set.
'' Read the DSP buffer as soon as data is available. The returned byte
'' should be &hAA for "reset successful."
''
'' Valid for SB1.5, SBMCV, SB2.0, SBPRO, SBPRO MCV, SB16
''
FUNCTION audDSPReset%
	DIM tmr AS SINGLE

	' No SoundBlaster configured, abort
	IF (sb.port = 0) THEN EXIT FUNCTION

	' Request DSP reset
	OUT sb.port + &H6, 1                              ' write 1 to reset port
	tmr = TIMER + .003: DO: LOOP WHILE (tmr > TIMER)  ' wait 3 msec
	OUT sb.port + &H6, 0                              ' write 0 to reset port

	' Check for confirmation
	tmr = TIMER + .1                                  ' allow 100 ms wait
	DO
		IF (INP(sb.port + &HE) AND 128) THEN            ' data is available
			audDSPReset% = (INP(sb.port + &HA) = &HAA)    ' test "reset" signal
			EXIT DO                                       ' we're done
		END IF
	LOOP WHILE (tmr > TIMER)
END FUNCTION

''
'' Commands go to the DSP buffer. Before sending a command, make sure
'' it is empty. Write-Buffer Status bit 7 is clear if the DSP buffer
'' is empty, set if there's something there.
''
'' Valid for SB1.5, SBMCV, SB2.0, SBPRO, SBPRO MCV, SB16
''
SUB audDSPWrite (anyByte AS INTEGER)
	DO: LOOP WHILE (INP(sb.port + &HC) AND 128) ' wait for "clear"
	OUT sb.port + &HC, anyByte                  ' write command
END SUB

''
'' Read from mixer chip
''
FUNCTION audMixerRead% (regIndex AS INTEGER)
	OUT sb.mixer + &H4, regIndex        ' Set register port
	audMixerRead% = INP(sb.mixer + &H5) ' Read data port
END FUNCTION

''
'' Write to mixer chip
''
SUB audMixerWrite (regIndex AS INTEGER, anyByte AS INTEGER)
	OUT sb.mixer + &H4, regIndex  ' Set register port
	OUT sb.mixer + &H5, anyByte   ' Write to data port
END SUB

''
'' Play wave buffer (MUST be 8 bit, 11025hz, mono, uncompressed)
'' Provide the segment, offset and length of the data buffer
''
SUB audPlayWav (wavSeg AS INTEGER, wavOfs AS INTEGER, wavLen AS LONG)
	DIM msg AS STRING * 6, wavAdr AS LONG, ofs AS INTEGER
	DIM DMAChannel AS INTEGER, DMATerminal AS INTEGER
	PRINT "Playing sound 1"
	' No SoundBlaster configured, abort
	IF (sb.port = 0) THEN EXIT SUB

	IF (wavLen < 1) THEN EXIT SUB
	PRINT "Playing sound sb.port:"; sb.port; " wavLen: "; wavLen

	' Force continue transfer if DMA transfer is not complete
	IF (audDMADone% = 0) THEN
		PRINT "Force continue transfer if DMA transfer is not complete"
		audDSPWrite &HD4
		PRINT "Force continue done"
	END IF

	' linear address and length of sound buffer
	IF (wavSeg < 0) THEN
		PRINT "wavSeg < 0"
		wavAdr = (CLNG(wavSeg) + 65536) * 16&
	ELSE
		wavAdr = CLNG(wavSeg) * 16&
	END IF
	IF (wavOfs < 0) THEN
		PRINT "wavOfs < 0"
		wavAdr = wavAdr + (wavOfs + 65536)
	ELSE
		wavAdr = wavAdr + wavOfs
	END IF
	msg = MKL$(wavAdr) + MKI$(wavLen - 1)

	' Here's the painful part: programming the DMA (Direct Memory
	' Access) controller. It's like a highway connecting memory and I/O
	' devices (like, say, a sound card.) This allows us to send huge
	' chunks of data without the help of the CPU. The idea here is to
	' configure the DMA controller to send our sample to the Sound
	' Blaster right away.

	' Bit on status register
	DMAChannel = (sb.loDMA AND &H3)
	'DMATerminal = 2 ^ DMAChannel
	PRINT "DMAChannel: "; DMAChannel
	' DMA mask port. We set bit 2 (value 0x04) to SET THE MASK BIT; that
	' is, we DISABLE the channel. Bits 1 and 0 are used to send the
	' channel number (that should be the 2 least significant bits of the
	' loDMA: 00 for channel 0 or 4, 01 for channel 1 or 5, 10 for
	' channel 2 or 6, 11 for channel 3 or 7.)
	OUT sb.DMAMask, 4 OR DMAChannel

	' DMA clear port. Send anything to clear the byte pointer flip-flop.
	PRINT "DMA Clear : sending port: "; sb.DMAClr; " data: 0"
	OUT sb.DMAClr, 0

	' DMA mode port. Used to select the transfer mode. Usually 0x48 for
	' single-cycle playback, 0x58 for auto-initialized playback, 0x44
	' for single-cycle record and 0x54 for auto-initialized recording.
	' The two least significant bits are reserved to send the channel
	' number (see DMA mask port.)
	' !!! THIS ONLY WORKS WITH DSP VERSION 2.xx+, shouldn't work on
	' SB1.0 or 1.5 !!!

	PRINT "DMA Mode: sending port: "; sb.DMAMode; " data: "; &H48 OR DMAChannel

	OUT sb.DMAMode, &H48 OR DMAChannel

	' DMA addres port and Provide the location and length of the data.
	' Note that the length of the data is minus 1 (we've done that when
	' creating the <msg> string already.)
	DEF SEG = VARSEG(msg)
	ofs = VARPTR(msg)
	PRINT "OUT addr: "; sb.DMAAddr; " value: "; ofs
	OUT sb.DMAAddr, PEEK(ofs)      ' Address (bits 0 to 7)

	PRINT "OUT addr: "; sb.DMAAddr; " value: "; ofs + 1
	OUT sb.DMAAddr, PEEK(ofs + 1)  ' Address (bits 8 to 15)
   
	PRINT "OUT addr: "; sb.DMAAddr; " value: "; ofs + 2
	OUT sb.DMAPage, PEEK(ofs + 2)  ' Page (bits 16 to 23 of address)
   
	PRINT "OUT addr: "; sb.DMAAddr; " value: "; ofs + 4
	OUT sb.DMASize, PEEK(ofs + 4)  ' Length (bits 0 to 7)
   
	PRINT "OUT addr: "; sb.DMAAddr; " value: "; ofs + 5
	OUT sb.DMASize, PEEK(ofs + 5)  ' Length (bits 8 to 15)
	DEF SEG

	' DMA mask port. Again. If bit 2 is clear, we CLEAR THE MASK BIT; so
	' we're ENABLING the channel whose index is provided in the two
	' least significant bits.

	PRINT "MASK OUT addr:  "; sb.DMAMask; " val: "; DMAChannel
	OUT sb.DMAMask, DMAChannel

	IF (sb.dspVer < 400) THEN

		PRINT "DspVer < 400: "; sb.dspVer; " : calling audDSPWrite &H40 & 165"

		' This should be done for older cards using time constant. This
		' could be computed once and stored in the waveData structure.
		'timeconst% = 256 - 1000000 / (source.freq * source.chan) ' Get time constant
		audDSPWrite &H40  ' "Transfer time constant" command
		audDSPWrite 165   ' Time constant = 256 - 1000000 / (sound frequency * sound channels)
	ELSE
		PRINT "DspVer >= 400: "; sb.dspVer; " : calling audDSPWrite &H41 & 43, 17"

		' This should work with Sound Blaster cards using DSP version
		' 4.xx (SB16 and every card that came after.) Note that the old
		' time constant command (0x40) is still supported, so this
		' method is not mendatory. Also, this is not supported by some
		' virtual machines (looking at you, VirtualBox.) It's on them.
		audDSPWrite &H41
		audDSPWrite 43 ' (sound frequency \ &H100)
		audDSPWrite 17 ' (sound frequency AND &HFF)
	END IF

	' Set I/O transfer mode to the DSP and then write the block size.
	'IF (sound frequency > 22000) THEN
		'audDSPWrite &H48   ' "Set block transfer size" command (generates interrupt)
	'ELSE

	PRINT "audDSPWrite &H14 being called"
	audDSPWrite &H14   ' "8-bit single-cycle DMA mode" output command
	'END IF

	DEF SEG = VARSEG(msg)
	ofs = VARPTR(msg)

	PRINT "audDSPWrite size in samples low byte: "; ofs + 4
	audDSPWrite PEEK(ofs + 4)      ' size in samples, low byte
   
	PRINT "audDSPWrite size in samples hight byte: "; ofs + 5
	audDSPWrite PEEK(ofs + 5)      ' size in samples, high byte
	DEF SEG

	'IF (sound frequency > 22000) THEN
		'audDSPWrite &H91   ' High Speed DMA mode 8-bit
	'END IF

	' This waits for the DMA transfer to be over. The bit for the
	' selected channel should be set in the DMA status port if the
	' transfer is over... somehow, that doesn't work in VirtualBox...
	'' ============= GOOD CHUNK OF DEBUG ============= ''
	' This will freeze VirtualBox because of its faulty DMA emulation
	'DO: LOOP UNTIL (INP(sb.DMAStat) AND DMATerminal)

	' This right here should be done by a custom ISR; read the DSP
	' Read-Buffer Status port once to acknowledge the DSP interrupt.
	'tmp = INP(sb.port + &HE)

	' Acknowledge the end of interrupt with the Programmable Interrupt
	' Controller (PIC); extra step required for IRQ 8-15.
	'OUT &H20, &H20
	'SELECT CASE sb.irq
		'CASE 8 TO 15
			'OUT &HA0, &H20
	'END SELECT
END SUB

''
'' Parses <msg> and collect SoundBlaster settings. Returns an error
'' code if something went wrong, 0 otherwise.
''
FUNCTION audSBInit% (msg AS STRING)
	DIM ofs AS INTEGER, wrd AS INTEGER, num AS STRING

	' Reset configuration
	sb.port = 0
	sb.irq = 0
	sb.loDMA = 0
	sb.HiDMA = 0
	sb.mixer = 0
	sb.mpu = 0
	sb.series = 0
	sb.dspVer = 0
	sb.chip = 0

	' Parse configuration string
	msg = msg + " "
	ofs = 1
	DO
		wrd = INSTR(ofs + 1, msg, " ")
		num = MID$(msg, ofs + 1, wrd - ofs - 1)
		SELECT CASE ASC(MID$(msg, ofs, 1))
			CASE 65, 97  ' Aa, Base port, in hexadecimal
				sb.port = VAL("&h" + num)
				sb.mixer = sb.port
			'CASE 69, 101 ' Ee, AWE 32 "High" DMA Channel
				' honestly, I don't know what to do with that info
			CASE 68, 100 ' Dd, Low DMA (8-bit)
				sb.loDMA = VAL(num)
			CASE 72, 104 ' Hh, High DMA (16-bit)
				sb.HiDMA = VAL(num)
			CASE 73, 105 ' Ii, Interrupt request (IRQ)
				sb.irq = VAL(num)
			CASE 77, 109 ' Mm, mixer chip base I/O port, in hexademical
				sb.mixer = VAL("&h" + num) ' Usually the base port...
			CASE 80, 112 ' Pp, MPU401 chipset port (MIDI,) in hexadecimal
				sb.mpu = VAL("&h" + num)
			CASE 84, 116 ' Tt, Card type
				sb.series = VAL(num)
		END SELECT
		ofs = wrd + 1
	LOOP WHILE (ofs < LEN(msg))

	' Early bail
	IF (sb.port = 0) THEN
		audSBInit% = sbErrNoBasePort
		EXIT FUNCTION
	END IF

	' Attempt to reset DSP
	IF NOT audDSPReset% THEN
		sb.port = 0
		audSBInit% = sbErrDSPFailedReset
		EXIT FUNCTION
	END IF

	' Guess card type
	IF (sb.series = 0) THEN
		SELECT CASE sb.port
			CASE &H210, &H230
				sb.series = cardSB1
			CASE &H220, &H240
				sb.series = cardSB2
			CASE &H250, &H260
				IF (sb.HiDMA) THEN
					sb.series = cardSB16
				ELSE
					sb.series = cardSB2
				END IF
			CASE ELSE
				IF (sb.HiDMA) THEN
					sb.series = cardSB16
				ELSE
					sb.series = cardSBPro2
				END IF
		END SELECT
	END IF

	' Guess mixer chipset type
	SELECT CASE sb.series
		CASE cardSB1
			sb.chip = chipCT1335
		CASE cardSB2
			SELECT CASE sb.port
				CASE &H220, &H240
					' no chipset (SB 2.0)
				CASE ELSE
					sb.chip = chipCT1335 ' SB 2.0 CD
			END SELECT
		CASE cardSBPro, cardSBPro2, cardSBPro2b
			sb.chip = chipCT1345 ' supports low-pass filter output and stereo output
		CASE cardSB16
			sb.chip = chipCT1745
	END SELECT

	' Get DSP version (send 0xE1 to write command/data port)
	audDSPWrite &HE1  ' send "Get DSP Version" command 0xE1
	sb.dspVer = audDSPRead% * 100 + audDSPRead%

	' Reset mixer (function 0x00, send any byte)
	audMixerWrite &H0, 0

	' Get Direct Memory Access (DMA) ports
	SELECT CASE sb.loDMA
		CASE 0 TO 3
			sb.DMAMask = &HA  ' DMAC1, 8-bit: write single mask regiser
			sb.DMAMode = &HB  ' DMAC1, 8-bit: write mode register
			sb.DMAClr = &HC   ' DMAC1, 8-bit: clear byte pointer flip-flop
			sb.DMAStat = &H8  ' DMAC1, 8-bit: control/status registers
			sb.DMAAddr = sb.loDMA * 2
			sb.DMASize = sb.DMAAddr + 1
		CASE IS > 3
			sb.DMAMask = &HD4 ' DMAC2, 16-bit: write single mask regiser
			sb.DMAMode = &HD6 ' DMAC2, 16-bit: write mode register
			sb.DMAClr = &HD8  ' DMAC2, 16-bit: clear byte pointer flip-flop
			sb.DMAStat = &HD0 ' DMAC1, 16-bit: control/status registers
			sb.DMAAddr = 192 + (sb.loDMA - 4) * 4
			sb.DMASize = sb.DMAAddr + 2
	END SELECT
	SELECT CASE sb.loDMA
		CASE 0
			sb.DMAPage = &H87 ' 8-bit DMA channel 0 page
		CASE 1
			sb.DMAPage = &H83 ' 8-bit DMA channel 1 page
		CASE 2
			sb.DMAPage = &H81 ' 8-bit DMA channel 2 page
		CASE 3
			sb.DMAPage = &H82 ' 8-bit DMA channel 3 page
		CASE 4
			sb.DMAPage = &H8F ' should be 16-bit DMA channel 4 page?
		CASE 5
			sb.DMAPage = &H8B ' 16-bit DMA channel 5 page
		CASE 6
			sb.DMAPage = &H89 ' 16-bit DMA channel 6 page
		CASE 7
			sb.DMAPage = &H8A ' 16-bit DMA channel 7 page
	END SELECT
END FUNCTION

''
'' Turn speaker on or off
''
SUB audSpeaker (enable AS INTEGER)
	' No SoundBlaster configured, abort
	IF (sb.port = 0) THEN EXIT SUB

	IF (enable) THEN
		audDSPWrite &HD1 ' turn speaker on
	ELSE
		IF (audDMADone% = 0) THEN audDSPWrite &HD0 ' pause 8-bit DMA mode digitized sound I/O (SB16)
		audDSPWrite &HD3 ' turn speaker off
	END IF
END SUB

''
'' Get BASS, MASTER, MIDI, TREBLE and VOICE level (range 0-31)
''
SUB audVolGet (mode AS INTEGER, lf AS INTEGER, rt AS INTEGER)
	' Reset return values
	lf = 0: rt = 0

	' No SoundBlaster configured, abort
	IF (sb.port = 0) THEN EXIT SUB

	SELECT CASE mode
		' Bass control is unsupported by chips CT1335 and CT1345. The left
		' and right levels are stored on 4 bits (bit 7-4,) but rescaled to
		' 5 bits (in range 0 to 31) just to normalize the routine input
		' values.
		CASE volBASS
			SELECT CASE sb.chip
				CASE chipCT1745
					lf = ((audMixerRead%(&H46) AND 240) \ 16) * 2
					rt = ((audMixerRead%(&H47) AND 240) \ 16) * 2
			END SELECT

		' Master volume is obtained via the standard mixer read method
		CASE volMASTER
			SELECT CASE sb.chip
				CASE chipCT1335
					audVolGetGen &H2, lf, rt
				CASE chipCT1345
					audVolGetGen &H22, lf, rt
				CASE chipCT1745
					audVolGetGen &H30, lf, rt
			END SELECT

		' MIDI volume is obtained via the standard mixer read method
		CASE volMIDI
			SELECT CASE sb.chip
				CASE chipCT1335
					audVolGetGen &H6, lf, rt
				CASE chipCT1345
					audVolGetGen &H26, lf, rt
				CASE chipCT1745
					audVolGetGen &H34, lf, rt
			END SELECT

		' CD volume is obtained via the standard mixer read method
		CASE volCD
			SELECT CASE sb.chip
				CASE chipCT1335
					audVolGetGen &H8, lf, rt
				CASE chipCT1345
					audVolGetGen &H28, lf, rt
				CASE chipCT1745
					audVolGetGen &H36, lf, rt
		END SELECT

		' Treble control is unsupported by chips CT1335 and CT1345. The
		' left and right levels are stored on 4 bits (bit 7-4,) but
		' rescaled to 5 bits (in range 0 to 31) just to normalize the
		' routine input values.
		CASE volTREBLE
			SELECT CASE sb.chip
				CASE chipCT1745
					lf = ((audMixerRead%(&H44) AND 240) \ 16) * 2
					rt = ((audMixerRead%(&H45) AND 240) \ 16) * 2
			END SELECT

		' Voice volume is obtained via the standard method for mixer chips
		' CT1345 and CT1745. Chipset CT1335 is MONO (so left and right
		' channels have the same value) and the resolution is only 2 bits
		' (bits 2-1) and rescaled to 5 bits.
		CASE volVOICE
			SELECT CASE sb.chip
				CASE chipCT1335
					lf = ((audMixerRead%(&HA) AND 6) \ 2) * 8
					rt = lf
				CASE chipCT1345
					audVolGetGen &H4, lf, rt
				CASE chipCT1745
					audVolGetGen &H32, lf, rt
			END SELECT
	END SELECT
END SUB

''
'' Get generic MASTER, MIDI and VOICE volume
''
SUB audVolGetGen (regIndex AS INTEGER, lf AS INTEGER, rt AS INTEGER)
	DIM temp AS INTEGER

	SELECT CASE sb.chip
		' Volume is stored on 3 bits on the CT1335. The mixer is mono, so
		' the value for the left and right channel is identical. Volume is
		' stored on bits 3-1, then rescaled to 5 bits.
		CASE chipCT1335
			lf = ((audMixerRead%(regIndex) \ 2) AND 7) * 4
			rt = lf

		' CT1345 is stereo, so two values are read. Each value is 3 bits,
		' located at bit 3-1 (left channel) and 7-5 (right channel). Value
		' is rescaled to 5 bits.
		CASE chipCT1345
			temp = audMixerRead%(regIndex)
			lf = ((temp \ 2) AND 7) * 4
			rt = ((temp \ 32) AND 7) * 4

		' CT1745 is stereo and offers a higher resolution: the volume is
		' stored on 5 bits (bits 7-3.)
		CASE chipCT1745
			lf = (audMixerRead%(regIndex) \ 8) AND 31
			rt = (audMixerRead%(regIndex + 1) \ 8) AND 31
	END SELECT
END SUB

''
'' Set BASS, MASTER, MIDI, TREBLE and VOICE level
''
SUB audVolSet (mode AS INTEGER, lf AS INTEGER, rt AS INTEGER)
	DIM temp AS INTEGER, lc AS INTEGER, rc AS INTEGER

	' No SoundBlaster configured, abort
	IF (sb.port = 0) THEN EXIT SUB

	' Cap left value (5-bit value)
	SELECT CASE lf
		CASE IS < 0
			lc = 0
		CASE IS > 31
			lc = 31
		CASE ELSE
			lc = lf
	END SELECT

	' Cap right value (5-bit value)
	SELECT CASE rt
		CASE IS < 0
			rc = 0
		CASE IS > 31
			rc = 31
		CASE ELSE
			rc = rt
	END SELECT

	SELECT CASE mode
		' BASS volume only supported by CT1745. The volume is stored on
		' 4 bits (bits 7-4.) The value passed to audVolSet is 5-bit long,
		' so it is converted first. When writing to the mixer, bits 3-0
		' are preserved.
		CASE volBASS
			SELECT CASE sb.chip
				CASE chipCT1745
					temp = (audMixerRead%(&H46) AND 15) OR ((lc \ 2) * 16)
					audMixerWrite &H46, temp
					temp = (audMixerRead%(&H47) AND 15) OR ((rc \ 2) * 16)
					audMixerWrite &H47, temp
			END SELECT

		' Set master volume via standard method.
		CASE volMASTER
			SELECT CASE sb.chip
				CASE chipCT1335
					audVolSetGen &H2, lc, rc
				CASE chipCT1345
					audVolSetGen &H22, lc, rc
				CASE chipCT1745
					audVolSetGen &H30, lc, rc
			END SELECT

		' Set MIDI volume via standard method.
		CASE volMIDI
			SELECT CASE sb.chip
				CASE chipCT1335
					audVolSetGen &H6, lc, rc
				CASE chipCT1345
					audVolSetGen &H26, lc, rc
				CASE chipCT1745
					audVolSetGen &H34, lc, rc
			END SELECT

		' Set CD volume via standard method.
		CASE volCD
			SELECT CASE sb.chip
				CASE chipCT1335
					audVolSetGen &H8, lc, rc
				CASE chipCT1345
					audVolSetGen &H28, lc, rc
				CASE chipCT1745
					audVolSetGen &H36, lc, rc
		END SELECT

		' Treble volume only supported by CT1745. The volume is stored on
		' 4 bits (bits 7-4.) The value passed to audVolSet is 5-bit long,
		' so it is converted first. When writing to the mixer, bits 3-0
		' are preserved.
		CASE volTREBLE
			SELECT CASE sb.chip
				CASE chipCT1745
					temp = (audMixerRead%(&H44) AND 15) OR ((lc \ 2) * 16)
					audMixerWrite &H44, temp
					temp = (audMixerRead%(&H45) AND 15) OR ((rc \ 2) * 16)
					audMixerWrite &H45, temp
			END SELECT

		' Set voice volume via standard method for CT1345 and CT1745. For
		' chipset CT1335, use the value provided for the left channel.
		' Volume is stored on 2 bits (1-0,) bits 7-2 are represerved.
		CASE volVOICE
			SELECT CASE sb.chip
				CASE chipCT1335
					temp = (audMixerRead%(&HA) AND 252) OR (lc \ 8)
					audMixerWrite &HA, temp
				CASE chipCT1345
					audVolSetGen &H4, lc, rc
				CASE chipCT1745
					audVolSetGen &H32, lc, rc
			END SELECT
	END SELECT
END SUB

''
'' Set generic MASTER, MIDI and VOICE volume (MUST be in range 0-31)
''
SUB audVolSetGen (regIndex AS INTEGER, lf AS INTEGER, rt AS INTEGER)
	DIM temp AS INTEGER

	SELECT CASE sb.chip
		' Recale LF to 3 bits, preserve bits 0 and 7-4, replace 3-1.
		CASE chipCT1335
			temp = (audMixerRead%(regIndex) AND 241) OR ((lf \ 4) * 2)
			audMixerWrite regIndex, temp

		' Rescale LF and RT to 3 bits, preserve bits 0 and 4, replace bits
		' 3-1 (left channel) and 7-5 (right channel.)
		CASE chipCT1345
			temp = (audMixerRead%(regIndex) AND 238) OR ((lf \ 4) * 2) OR ((rf \ 4) * 32)
			audMixerWrite regIndex, temp

		' Preserve bits 2-0, replace bits 7-3
		CASE chipCT1745
			temp = (audMixerRead%(regIndex) AND 7) OR (lf * 8)
			audMixerWrite regIndex, temp
			temp = (audMixerRead%(regIndex + 1) AND 7) OR (rt * 8)
			audMixerWrite regIndex + 1, temp
	END SELECT
END SUB



DECLARE SUB memFree (segment AS INTEGER)
DECLARE FUNCTION memAlloc% (numBytes AS LONG)
DECLARE FUNCTION loadWAVE% (filename AS STRING, target AS ANY)

TYPE waveData
	segment AS INTEGER  ' 000 ..2   Pointer to data in base memory
	length AS LONG      ' 002 ..4   Length in bytes
END TYPE              ' 6 BYTES

DIM sndFile AS waveData

' Initialize card using the environment string
DIM tmp AS STRING
tmp = ENVIRON$("BLASTER")
IF (LEN(tmp) = 0) THEN tmp = "A220 I5 D1 H5"
SELECT CASE audSBInit%(tmp)
	CASE sbErrNoBasePort
		PRINT "audSBInit: SoundBlaster base port undefined"
		END
	CASE sbErrDSPFailedReset
		PRINT "audSBInit: Failed to reset DSP"
		END
END SELECT


' load sound file
PRINT "Loading wave file (this will take a while)..."
IF NOT loadWAVE%("z_gib.wav", sndFile) THEN PRINT "Something went wrong while loading file": END
audVolSet volMASTER, 32, 32

PRINT
PRINT " Segment: " + HEX$(sndFile.segment)
PRINT "  Length:"; sndFile.length
PRINT

PRINT "Space to play sound, escape to exit."
audSpeaker -1
DO
	r$ = INKEY$
	IF LEN(r$) THEN
		SELECT CASE ASC(r$)
			CASE 32
				audPlayWav sndFile.segment, 0, sndFile.length
				PRINT "Playing sound"
			CASE 27
				EXIT DO
		END SELECT
	END IF
LOOP
audSpeaker 0
memFree sndFile.segment

FUNCTION loadWAVE% (filename AS STRING, target AS waveData)
	DIM ff AS INTEGER

	DIM ckID AS LONG, ckSize AS LONG, formType AS LONG, sckID1 AS LONG, sckSize1 AS LONG
	DIM wFormatTag AS INTEGER, wChannels AS INTEGER, dwSamplesPerSec AS LONG
	DIM dwAvgBytesPerSec AS LONG, wBlockAlign AS INTEGER, wBitsPerSample AS INTEGER
	DIM sckID2 AS LONG, sckSize2 AS LONG

	DIM remain AS LONG, segment AS LONG, msg AS STRING

	ff = FREEFILE
	OPEN filename FOR BINARY AS #ff
	GET #ff, , ckID
	GET #ff, , ckSize
	IF (ckID = &H46464952) THEN ' "RIFF"
		IF (ckSize = LOF(ff) - 8) THEN
			GET #ff, , formType
			IF (formType = &H45564157) THEN ' "WAVE"
				GET #ff, , sckID1
				IF (sckID1 = &H20746D66) THEN ' "fmt "
					GET #ff, , sckSize1
					GET #ff, , wFormatTag
					GET #ff, , wChannels
					IF (wChannels = 1) THEN
						GET #ff, , dwSamplesPerSec
						IF (dwSamplesPerSec = 11025) THEN
							GET #ff, , dwAvgBytesPerSec
							GET #ff, , wBlockAlign
							GET #ff, , wBitsPerSample
							SEEK #ff, SEEK(ff) + (sckSize1 - 16)
							IF (wBitsPerSample = 8) THEN
								GET #ff, , sckID2
								IF (sckID2 = &H61746164) THEN ' "data"
									GET #ff, , sckSize2
									target.length = sckSize2
									target.segment = memAlloc%(target.length)
									IF (target.segment) THEN

										' That's VERY slow...
										segment = target.segment
										IF (segment < 0) THEN segment = segment + 65536
										remain = target.length
										msg = SPACE$(1024)
										DO
											IF (remain < 1024) THEN msg = SPACE$(remain)
											GET #ff, , msg
											DEF SEG = segment
											FOR i% = 0 TO LEN(msg) - 1
												POKE i%, ASC(MID$(msg, i% + 1, 1))
											NEXT i%
											segment = segment + 64
											remain = remain - LEN(msg)
										LOOP WHILE remain
										DEF SEG

										loadWAVE% = -1
									END IF
								END IF
							END IF
						END IF
					END IF
				END IF
			END IF
		END IF
	END IF
	CLOSE #ff
END FUNCTION

''
'' QuickBASIC moves stuff around in memory so it's not possible to
'' keep a static pointer; when it starts, QuickBASIC reserves the
'' largest block of memory available for the far heap. We need to free
'' some of that memory and request it back via DOS interrupt 0x21 to
'' take it off QuickBASIC's control. Requires DOS 2+
''
FUNCTION memAlloc% (numBytes AS LONG)
	DIM memReq AS LONG, junk AS LONG, regs AS RegTypeX

	IF numBytes THEN
		' Blocks are groups of 16 bytes.
		memReq = (numBytes \ 16) - ((numBytes AND 15) > 0)

		' Tell QuickBASIC to free some memory. A margin is needed but I
		' don't know why.
		junk = SETMEM(-(memReq + 1) * 16)

		' Use DOS Interrupt 0x21, function 0x48 to request that memory.
		regs.ax = &H4800
		regs.bx = memReq
		CALL INTERRUPTX(&H21, regs, regs)

		' If CF is set, something went wrong. Restore memory.
		IF (regs.flags AND &H1) THEN
			junk = SETMEM(650000)
			memAlloc% = 0
		' If CF is clear, AX contains the segment address to the first
		' paragraph of the reserved block.
		ELSE
			memAlloc% = regs.ax
		END IF
	END IF
END FUNCTION

''
'' Free memory reserved via memAlloc%(). Requires DOS 2+
''
SUB memFree (segment AS INTEGER)
	DIM junk AS LONG, regs AS RegTypeX

	IF (segment) THEN
		' Free allocated memory.
		regs.ax = &H4900
		regs.es = segment
		CALL INTERRUPTX(&H21, regs, regs)

		' Clear segment and offset.
		segment = 0

		' Give back memory to QuickBASIC.
		junk = SETMEM(650000)
	END IF
END SUB

